package com.tzld.piaoquan.recommend.server.service.recall.strategy;

import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.tzld.piaoquan.recommend.server.service.filter.FilterParam;
import com.tzld.piaoquan.recommend.server.service.filter.FilterResult;
import com.tzld.piaoquan.recommend.server.service.filter.FilterService;
import com.tzld.piaoquan.recommend.server.service.recall.RecallParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.recall.RecallStrategy;
import com.tzld.piaoquan.recommend.server.util.DateUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ZSetOperations;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
public abstract class AbstractRegionRecallStrategy implements RecallStrategy {
    @Autowired
    protected RedisTemplate<String, String> redisTemplate;

    @Autowired
    protected FilterService filterService;

    @Override
    public RecallResult recall(RecallParam param) {

        String recordKey = recordKey(param);
        String lastVideoKey = lastVideoKey(param);
        int idx = 0;
        String poolKey;

        // TODO 整段逻辑很奇怪，需要重新理解
        if (redisTemplate.hasKey(recordKey)) {
            String record = redisTemplate.opsForValue().get("record_key");
            Map<String, String> recordMap = JSONObject.parseObject(record, Map.class);

            String record_dt = recordMap.get("date");
            String record_h = recordMap.get("h");
            String now_dt = DateUtils.getCurrentDateStr("yyyyMMdd");
            int h = DateUtils.getCurrentHour();
            if (record_dt.equals(now_dt) && Integer.valueOf(record_h) == h) {
                poolKey = poolKey(param, now_dt, h);
                idx = getIdx(lastVideoKey, poolKey);
            } else if ((record_dt.equals(now_dt) && h - Integer.valueOf(record_h) == 1)
                    || (h == 0 && Integer.valueOf(record_h) == 23)) {
                poolKey = poolKey(param, now_dt, h);
                if (redisTemplate.hasKey(poolKey)) {
                    redisTemplate.opsForValue().set(recordKey, String.format("{\"date\":%s, \"h\":%d}", now_dt, h), 2,
                            TimeUnit.HOURS);
                    redisTemplate.delete(lastVideoKey);
                } else {
                    poolKey = poolKey(param, record_dt, Integer.valueOf(record_h));
                    idx = getIdx(lastVideoKey, poolKey);
                }
            } else {
                redisTemplate.delete(lastVideoKey);
                poolKey = updateLastVideoRecord(recordKey, param);
            }
        } else {
            redisTemplate.delete(lastVideoKey);
            poolKey = updateLastVideoRecord(recordKey, param);
        }

        if (StringUtils.isBlank(poolKey)) {
            return null;
        }

        // 召回
        int getSize = param.getSize() * 5;
        int freq = 0;
        List<RecallResult.RecallData> results = new ArrayList<>();
        while (results.size() < param.getSize()) {
            freq += 1;
            if (freq > 2) {
                break;
            }
            Set<ZSetOperations.TypedTuple<String>> data = redisTemplate.opsForZSet().reverseRangeWithScores(poolKey, idx,
                    idx + getSize - 1);
            if (CollectionUtils.isEmpty(data)) {
                break;
            }
            idx += getSize;
            Map<Long, Double> videoMap = data.stream()
                    .filter(t -> NumberUtils.isDigits(t.getValue()))
                    .collect(Collectors.toMap(
                            t -> Long.getLong(t.getValue()),
                            t -> t.getScore(),
                            (a, b) -> b
                    ));
            FilterParam filterParam = new FilterParam();
            filterParam.setVideoIds(Lists.newArrayList(videoMap.keySet()));
            FilterResult filterResult = filterService.filter(filterParam);
            if (filterResult != null && CollectionUtils.isNotEmpty(filterResult.getVideoIds())) {
                filterResult.getVideoIds().stream().forEach(vid -> {
                    RecallResult.RecallData recallData = new RecallResult.RecallData();
                    recallData.setVideoId(vid);
                    recallData.setAbCode(param.getAbCode());
                    recallData.setRovScore(videoMap.get(vid));
                    recallData.setPushFrom(pushFrom());
                    results.add(recallData);
                });
            }
        }


//        pool_recall_result.sort(key=lambda x: x.get('rovScore', 0), reverse=True)
//
//        # if len(recall_data) > 0 and len(pool_recall_result) == 0 \
//        #         and self.ab_code == config_.AB_CODE['region_rank_by_h'].get('abtest_112') and self.mid:
//
//        if len(recall_data) > 0 and len(pool_recall_result) == 0 and self.mid:
//            # 召回数据不为空 & 过滤后结果为空 & 位于实验组 & mid不为空时，更新召回获取的末位视频id记录到定位的key中
//                last_video_key = f'{last_video_key_prefix}{self.app_type}:{self.mid}'
//        self.redis_helper.set_data_to_redis(key_name=last_video_key, value=recall_data[-1][0],
//                expire_time=expire_time)

        RecallResult result = new RecallResult();
        result.setRecallData(results.subList(0, results.size() < param.getSize() ? results.size() : param.getSize()));
        return result;
    }

    private int getIdx(String lastVideoKey, String poolKey) {
        int idx = 0;
        String lastVideo = redisTemplate.opsForValue().get(lastVideoKey);
        if (StringUtils.isNotBlank(lastVideo)) {
            Long idxLong = redisTemplate.opsForZSet().reverseRank(poolKey, lastVideo);
            if (idxLong != null) {
                idx = idxLong.intValue();
                idx += 1;
            }
        }
        return idx;
    }

    private String updateLastVideoRecord(String recordKey, RecallParam param) {
        String currentDateStr = DateUtils.getCurrentDateStr("yyyyMMdd");
        int h = DateUtils.getCurrentHour();

        String poolKey = poolKey(param, currentDateStr, h);
        if (!redisTemplate.hasKey(poolKey)) {
            if (h == 0) {
                currentDateStr = DateUtils.getBeforeDaysDateStr("yyyyMMdd", 1);
                h = 23;
            } else {
                h = h - 1;
            }
            poolKey = poolKey(param, currentDateStr, h);
        }
        redisTemplate.opsForValue().set(recordKey, String.format("{\"date\":%s, \"h\":%d}", currentDateStr, h), 2,
                TimeUnit.HOURS);
        return poolKey;
    }

    abstract String recordKey(RecallParam param);

    abstract String poolKey(RecallParam param, String now_dt, int h);

    abstract String lastVideoKey(RecallParam param);

    abstract String pushFrom();
}
